{--------------------------------------------------
Projekt:  Allgemein
Datei:    MmUtilK
Funktion: Allgemeine Funktionen, die oft benötigt
      werden.
---------------------------------------------------
 31.07.00 Mm  Unit Erstellt...
--------------------------------------------------}
{$mode objfpc}{$H+}

unit MmUtilK;

interface

uses Forms, SysUtils, Classes, FileUtil, strutils,
  Controls, Graphics, UTF8Process
  {$ifdef WINDOWS}, Windows{$endif}
  {$ifdef UNIX}, LCLIntf{$endif}
  //   {$ifdef WINDOWS}, Registry{$endif}     DBGrids, Translations, ZAbstractDataset, IniFiles, Math, Messages,
  ;

type
  TFileStreamUTF8 = class(THandleStream)
  private
    FFileName: string;
  public
    constructor Create(const AFileName: string; Mode: word; Rights: cardinal = 438);
    destructor Destroy; override;
    property FileName: string read FFilename;
  end;

  TStringListUTF8 = class(TStringList)
  public
    constructor Create;
    destructor Destroy; override;
    procedure SaveToFileUTF8(const FileName: string);
    procedure LoadFromFileUTF8(const FileName: string);
    function IndexOfText(const sVal: string): integer;
  end;

function FileCreateUTF8Mm(const FileName: string; ShareMode: integer = fmShareExclusive;
  Rights: integer = 0): THandle;
procedure ShellExec(Filename, Params: string);
function  OpenURLMm(AURL: String): Boolean;
function  OpenDocumentMm(APath: String): Boolean;
function DirectoryExistsUTF8Mm(const Directory: string): boolean;
function CreateDirUTF8Mm(const NewDir: string): boolean;
function DeleteFileUTF8Mm(const FileName: string): boolean;
function RenameFileUTF8Mm(const OldName, NewName: string): boolean;
function FindFirstUTF8Mm(const Path: string; Attr: longint;
  out Rslt: TSearchRec): longint;
function FindNextUTF8Mm(var Rslt: TSearchRec): longint;
function FileAgeUTF8Mm(const FileName: string): longint;
function FileSetAttrUTF8Mm(const FileName: string; Attr: longint): longint;
function FileSetDateUTF8Mm(const FileName: string; Age: longint): longint;
function FileOpenUTF8Mm(const FileName: string; Mode: integer): THandle;
function ParamStrUTF8Mm(Param: integer): string;
function CheckFilename(s: string; bAcceptPathDelim: Boolean): string;
function IsFileName(FileName: string): boolean;
function ShiftStateToString(Shift: TShiftState): String;

function ExpandEnvironmentString(s: string): string;
// Wandelt Linux Slash zu Windows Backslash je nach Betriebssystem
procedure ChangePathDelim(var s: string);
// Korrektur Pfade, Bildung Absoluter Pfad bei relativer Eingabe
function KorrectPath(s: string): string;
procedure CreateDirsUTF8(s: string);
function CopyFileUTF8(const SrcFilename, DestFilename: string;
  PreserveTime: boolean = False): boolean;

resourcestring
  SFCreateError = 'Unable to create file "%s"';
  SFOpenError = 'Unable to open file "%s"';

implementation

function FileOpenUTF8Mm(const FileName: string; Mode: integer): THandle;
{$ifdef WINDOWS}
const
  AccessMode: array[0..2] of cardinal =
    (GENERIC_READ, GENERIC_WRITE, GENERIC_READ or GENERIC_WRITE);
  ShareModes: array[0..4] of integer =
    (0, 0, FILE_SHARE_READ, FILE_SHARE_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE);
{$endif}
begin
{$ifdef WINDOWS}
  Result := CreateFileW(PWideChar(UTF8Decode(FileName)),
    dword(AccessMode[Mode and 3]), dword(
    ShareModes[(Mode and $F0) shr 4]), nil, OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL, 0);
  //if fail api return feInvalidHandle (INVALIDE_HANDLE=feInvalidHandle=-1)
{$else}
  Result := FileOpen(FileName, Mode);
{$endif}
end;

function FileCreateUTF8Mm(const FileName: string; ShareMode: integer = fmShareExclusive;
  Rights: integer = 0): THandle;
{$ifdef WINDOWS}
const
  //  AccessMode: array[0..2] of Cardinal  = (GENERIC_READ, GENERIC_WRITE, GENERIC_READ or GENERIC_WRITE);
  ShareModes: array[0..4] of integer =
    (0, 0, FILE_SHARE_READ, FILE_SHARE_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE);
{$endif}
begin
{$ifdef WINDOWS}
  Result := CreateFileW(PWideChar(UTF8Decode(FileName)), GENERIC_READ or
    GENERIC_WRITE, dword(ShareModes[(ShareMode and $F0) shr 4]),
    nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
{$else}
  Result := FileCreate(FileName, ShareMode, Rights);
{$endif}
end;

{****************************************************************************}
{*                             TFileStreamUTF8                              *}
{****************************************************************************}
constructor TFileStreamUTF8.Create(const AFileName: string; Mode: word;
  Rights: cardinal = 438);
var
  hFile: THandle;
begin
  FFileName := AFileName;
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    if (Mode and fmCreate) = fmCreate then
    begin
      if FileExists(AFileName) then
        DeleteFile(PChar(AFileName));
      hFile := FileCreate(AFileName, Mode, Rights);
    end
    else
      hFile := FileOpen(AFileName, Mode);
  end
  else
  {$endif}
  begin
    if (Mode and fmCreate) = fmCreate then
    begin
      if FileExistsUTF8(AFileName) then
        DeleteFileUTF8(AFileName);
      hFile := FileCreateUTF8Mm(AFileName, Mode, Rights);
    end
    else
      hFile := FileOpenUTF8Mm(AFileName, Mode);
  end;

  if hFile = THandle(-1) then
    if (Mode and fmCreate) = fmCreate then
      raise EFCreateError.createfmt(SFCreateError, [AFileName])
    else
      raise EFOpenError.Createfmt(SFOpenError, [AFilename]);
  inherited Create(hFile);
end;

destructor TFileStreamUTF8.Destroy;
begin
  FileClose(Handle);
  inherited Destroy;
end;


{****************************************************************************}
{*                             TFileStreamUTF8                              *}
{****************************************************************************}
constructor TStringListUTF8.Create;
begin
  inherited Create;
end;

destructor TStringListUTF8.Destroy;
begin
  inherited Destroy;
end;

procedure TStringListUTF8.SaveToFileUTF8(const FileName: string);
var
  TheStream: TFileStreamUTF8;
begin
  TheStream := TFileStreamUTF8.Create(FileName, fmCreate);
  try
    SaveToStream(TheStream);
  finally
    TheStream.Free;
  end;
end;

procedure TStringListUTF8.LoadFromFileUTF8(const FileName: string);
var
  TheStream: TFileStreamUTF8;
begin
  TheStream := TFileStreamUTF8.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(TheStream);
  finally
    TheStream.Free;
  end;
end;

function TStringListUTF8.IndexOfText(const sVal: string): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 0 to Count - 1 do
  begin
    if SameText(sVal, Self[i]) then
    begin
      Result := i;
      Break;
    end;
  end;
end;

function StrToFloatDefMm(sVal: String; fDefault: Double): Double;
Begin
  sVal := StringReplace(sVal, '.', DefaultFormatSettings.DecimalSeparator, []);
  sVal := StringReplace(sVal, ',', DefaultFormatSettings.DecimalSeparator, []);
  Result := StrToFloatDef(sVal, fDefault);
end;

procedure ShellExec(Filename, Params: string);
var
  TP: TProcessUTF8;
begin
  TP := TProcessUTF8.Create(nil);
  {$ifdef WINDOWS}
  TP.Executable := Filename;
  TP.Parameters.Clear;
  TP.Parameters.DelimitedText := Params;
  {$else}
  TP.CommandLine := Filename + ' ' + Params;
  {$endif}
  TP.Options := [];
  TP.Execute;
  TP.Free;
end;

{$ifdef UNIX}
Function OpenURLMm(AURL: String): Boolean;
Begin
  Result := OpenURL(AURL);
end;
{$ELSE}
function OpenURLMm(AURL: String): Boolean;
var
{$IFDEF WinCE}
  Info: SHELLEXECUTEINFO;
{$ELSE}
  PURL: String;
  ws, wsp: WideString;
  ans, ansp: AnsiString;
  IsFileUriWithSpaces: Boolean;
const
  FileURIScheme = 'file://';
{$ENDIF}
begin
  Result := False;
  if AURL = '' then Exit;
  PURL := ExtractFileDir(AURL);
  If PURL = AURL Then PURL := '';
  If Not DirectoryExistsUTF8Mm(PURL) Then PURL := '';
  {$IFDEF WinCE}
  FillChar(Info, SizeOf(Info), 0);
  Info.cbSize := SizeOf(Info);
  Info.fMask := SEE_MASK_FLAG_NO_UI;
  Info.lpVerb := 'open';
  Info.lpFile := PWideChar(UTF8Decode(AURL));
  ToDo:
  Info.lpDir := ....
  Result := ShellExecuteEx(@Info);
  {$ELSE}
  if Win32Platform = VER_PLATFORM_WIN32_NT then
  begin
    //Urls that start with file:// are allowed to contain spaces and should be quoted on NT platform,
    //but on Win9x quoting it fails
    //Since on Windows filenames cannot contain the " character, we need not care about it and simply enclose the AURL
    IsFileUriWithSpaces := (Pos(#32,AURL) > 0) and (CompareText(Copy(AURL,1,Length(FileURIScheme)), FileURIScheme) = 0);
    if IsFileUriWithSpaces then AURL := '"' + AURL + '"';
    ws := UTF8Decode(AURL);

    if PURL <> '' Then
    Begin
      IsFileUriWithSpaces := (Pos(#32,PURL) > 0) and (CompareText(Copy(PURL,1,Length(FileURIScheme)), FileURIScheme) = 0);
      if IsFileUriWithSpaces then PURL := '"' + PURL + '"';
      wsp := UTF8Decode(PURL);
    end Else wsp := '';
    Result := ShellExecuteW(0, nil, PWideChar(ws), nil, PWideChar(wsp), SW_SHOWNORMAL) > 32;
  end
  else
  begin
    ans := Utf8ToAnsi(AURL); // utf8 must be converted to Windows Ansi-codepage
    ansp := Utf8ToAnsi(PURL);
    Result := ShellExecute(0, nil, PAnsiChar(ans), nil, PAnsiChar(ansp), SW_SHOWNORMAL) > 32;
  end;
  {$ENDIF}
end;
{$ENDIF}

// Open a document with the default application associated with it in the system
{$ifdef UNIX}
Function OpenDocumentMm(APath: String): Boolean;
Begin
  Result := OpenDocument(APath);
end;
{$ELSE}
function OpenDocumentMm(APath: String): Boolean;
begin
  Result := OpenURLMm(APath);
end;
{$ENDIF}

function FileExistsUTF8Mm(sFileName: string): boolean;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FileExists(sFileName);
  end
  else
  {$endif}
    Result := FileExistsUTF8(sFileName);
end;

function FileGetAttrUTF8Mm(const FileName: string): longint;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FileGetAttr(FileName);
  end
  else
  {$endif}
    Result := FileGetAttrUTF8(FileName);
end;

function DirectoryExistsUTF8Mm(const Directory: string): boolean;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := DirectoryExists(Directory);
  end
  else
  {$endif}
    Result := DirectoryExistsUTF8(Directory);
end;

function CreateDirUTF8Mm(const NewDir: string): boolean;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := CreateDir(NewDir);
  end
  else
  {$endif}
    Result := CreateDirUTF8(NewDir);
end;

function ParamStrUTF8Mm(Param: integer): string;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := ParamStr(Param);
  end
  else
  {$endif}
    Result := ParamStrUTF8(Param);
end;

function DeleteFileUTF8Mm(const FileName: string): boolean;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := DeleteFile(PChar(FileName));
  end
  else
  {$endif}
    Result := DeleteFileUTF8(FileName);
end;

function RenameFileUTF8Mm(const OldName, NewName: string): boolean;
Begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := RenameFile(OldName, NewName);
  end
  else
  {$endif}
    Result := RenameFileUTF8(OldName, NewName);
End;

function FindFirstUTF8Mm(const Path: string; Attr: longint;
  out Rslt: TSearchRec): longint;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FindFirst(Path, Attr, Rslt);
  end
  else
  {$endif}
    Result := FindFirstUTF8(Path, Attr, Rslt);
end;

function FindNextUTF8Mm(var Rslt: TSearchRec): longint;
begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FindNext(Rslt);
  end
  else
  {$endif}
    Result := FindNextUTF8(Rslt);
end;

function FileAgeUTF8Mm(const FileName: string): longint;
Begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FileAge(FileName);
  end
  else
  {$endif}
    Result := FileAgeUTF8(FileName);
end;

function FileSetAttrUTF8Mm(const FileName: string; Attr: longint): longint;
Begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FileSetAttr(FileName, Attr);
  end
  else
  {$endif}
    Result := FileSetAttrUTF8(FileName, Attr);
end;

function FileSetDateUTF8Mm(const FileName: string; Age: longint): longint;
Begin
  {$ifdef WINDOWS}
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    Result := FileSetDate(FileName, Age);
  end
  else
  {$endif}
    Result := FileSetDateUTF8(FileName, Age);
end;

// Löscht alle unbeliebten Zeichen aus dem Dateiname
function CheckFilename(s: string; bAcceptPathDelim: Boolean): string;
const
  ch = [#0..#31, '*', '<', '>', '|', '?', '"', '[', ']', '+', ';', '=']; // '/', ':', '\',
var
  i: integer;
begin
  for i := Length(s) downto 1 do
    if s[i] in ch then
      Delete(s, i, 1);
  If Not bAcceptPathDelim Then
  Begin
    s := StringReplace(s, '\', '', [rfReplaceAll]);
    s := StringReplace(s, '/', '', [rfReplaceAll]);
    s := StringReplace(s, ':', '', [rfReplaceAll]);
  end;
  Result := s;
end;

////////////////////////////////////////////////////////////////////////////////
// Dateiname auf Gültigkeit überprüfen
////////////////////////////////////////////////////////////////////////////////
function IsFileName(FileName: string): boolean;
const
  ForbiddenChars = ['"', '<', '>', '|', '*', '/', '\', '?']; // verbotene Zeichen
const
  ForbiddenNames: array[0..22] of string[6] =
    ('AUX', 'NUL', 'PRN', 'CON', 'CLOCK$', // verbotene Namen
    'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', 'COM8', 'COM9',
    'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9');
var
  i: integer;
var
  p: PChar;
var
  FileNameU: string;
begin
  Result := False;
  if FileName <> '' then // Name darf nicht leer sein
  begin
    i := Length(FileName);
    if FileName[i] <> '.' then // letze Zeichen darf kein Punkt sein
    begin
     {$ifdef WINDOWS}
      p := PChar(ansistring(FileName));
     {$else}
      p := PChar(ansistring(FileName));
     {$endif}
      repeat
        if p^ in ForbiddenChars then
          Exit;
        Inc(p);
      until p^ = #0;
      if (i < 7) and (i > 2) then
      begin
        FileNameU := UpperCase(FileName);
        for i := 0 to High(ForbiddenNames) do
        begin
          if CompareStr(ForbiddenNames[i], FileNameU) = 0 then
            Exit;
        end;
      end;
      Result := True;
    end;
  end;
end;

function ShiftStateToString(Shift: TShiftState): String;
Var s: String;
Begin
  s := '';
  If ssShift In Shift Then s := 'ssShift';
  If ssAlt In Shift Then s := s + ' ssAlt';
  If ssCtrl In Shift Then s := s + ' ssCtrl';
  If ssLeft In Shift Then s := s + ' ssLeft';
  If ssRight In Shift Then s := s + ' ssRight';
  If ssMiddle In Shift Then s := s + ' ssMiddle';
  If ssDouble In Shift Then s := s + ' ssDouble';
  If ssMeta In Shift Then s := s + ' ssMeta';
  If ssSuper In Shift Then s := s + ' ssSuper';
  If ssHyper In Shift Then s := s + ' ssHyper';
  If ssAltGr In Shift Then s := s + ' ssAltGr';
  If ssCaps In Shift Then s := s + ' ssCaps';
  If ssNum In Shift Then s := s + ' ssNum';
  If ssScroll In Shift Then s := s + ' ssScroll';
  If ssTriple In Shift Then s := s + ' ssTriple';
  If ssQuad In Shift Then s := s + ' ssQuad';
  If ssExtra1 In Shift Then s := s + ' ssExtra1';
  If ssExtra2 In Shift Then s := s + ' ssExtra2';
  Result := s;
end;

function ExpandEnvironmentString(s: string): string;
var
  i, k: integer;
  sl: TStringListUTF8;
  s2: string;
  bP, bD: boolean;
begin
  Result := s;
  bP := Pos('%', s) > 0;
  bD := Pos('$', s) > 0;
  if not (bP or bD) then
    Exit; // Nichts zu konvertieren
  sl := TStringListUTF8.Create;
  sl.Sorted := bD;
  for i := GetEnvironmentVariableCount - 1 downto 0 do
    // Alle Variablen in eine Liste merken
  begin
    s2 := GetEnvironmentStringUTF8(i);
    if Length(s2) >= 2 then
      if s2[1] <> '=' then
        sl.Add(s2);
  end;

  k := 0;
  repeat // Durchlaufe so oft bis alle Env-Variablen aufgelöst sind
    s2 := Result;
    for i := sl.Count - 1 downto 0 do
    begin
      if bP then
      begin
        Result := StringReplace(Result, '%' + sl.Names[i] + '%',
          sl.ValueFromIndex[i], [rfReplaceAll, rfIgnoreCase]);
        if Pos('%', Result) = 0 then
          bP := False;
      end;
      if bD then
      begin
        Result := StringReplace(Result, '$' + sl.Names[i], sl.ValueFromIndex[i],
          [rfReplaceAll]);
        if Pos('$', Result) = 0 then
          bD := False;
      end;
      if not (bP or bD) then
        Break;
    end;
    if not (bP or bD) then
      Break;
    if SameText(s2, Result) then
      Break; // % $ nicht konvertierbar, da im System nicht vorhanden
    Inc(k);
    if k > sl.Count - 2 then
      Break; // Max Durchläufe
  until True;
  sl.Free;
end;

procedure ChangePathDelim(var s: string);
// Wandelt Linux Slash zu Windows Backslash je nach Betriebssystem
begin
  if SameText(PathDelim, '\') then
  begin
    if Pos('/', s) > 0 then // Linux-Pfade auf Windows umbiegen
      s := StringReplace(s, '/', PathDelim, [rfReplaceAll]);
  end
  else
  begin
    if Pos('\', s) > 0 then // Windws-Pfade auf Linux umbiegen
      s := StringReplace(s, '\', PathDelim, [rfReplaceAll]);
  end;
end;

function KorrectPath(s: string): string;
  // Korrektur Pfade, Bildung Absoluter Pfad bei relativer Eingabe
var
  sExe: string;
  iPos, iPos2: integer;
begin
  if Copy(s, Length(s), 1) = PathDelim then
    s := Copy(s, 1, Length(s) - 1); // Letztes \ löschen
  ChangePathDelim(s);
  {$ifdef Darwin}
  if (Copy(s, 1, 1) = '~') Then
    s := StringReplace(s, '~', GetUserDir, []);
  {$endif}
  if (Copy(s, 1, 2) = ('.' + PathDelim)) or (s = '.') then
  begin // Wenn Path-Parameter Relativ ist, dann auf Absolut ändern
    Delete(s, 1, 1);
    s := ExtractFileDir(Application.ExeName) + s;
  end;
  while Pos(PathDelim + '.' + PathDelim, s) > 0 do  // Alle .\ löschen
    Delete(s, Pos(PathDelim + '.' + PathDelim, s), 2);
  if Copy(s, 1, 3) = ('..' + PathDelim) then
  begin // Wenn Path-Parameter übergeordnet Relativ ist, dann auf Absolut ändern
    Delete(s, 1, 3);
    sEXE := ExtractFileDir(ExtractFileDir(Application.ExeName));
    while Pos('..' + PathDelim, s) = 1 do
    begin
      Delete(s, 1, 3);
      sEXE := ExtractFileDir(sExe);
    end;
    s := StringReplace(sExe + PathDelim + s, PathDelim + PathDelim, PathDelim, []);
  end;
  while Pos(PathDelim + '..' + PathDelim, s) > 1 do
    // Wechsel zum übergeordneten Dir ist noch drin
  begin
    iPos := Pos(PathDelim + '..' + PathDelim, s);
    iPos2 := RPosEX(PathDelim, s, iPos - 1);
    if iPos2 <= 0 then
    begin
      while iPos > 0 do
      begin
        Delete(s, iPos, 3);
        iPos := Pos(PathDelim + '..' + PathDelim, s);
      end;
      Break;
    end;
    Delete(s, iPos2, iPos - iPos2 + 3);
  end;
  if RPos(PathDelim + '.', s) = (Length(s) - 1) then // Zu letzt steht ein \.
    Delete(s, Length(s) - 1, 2);
  iPos := RPos(PathDelim + '..', s);
  if iPos = (Length(s) - 2) then // Zu letzt steht ein \..
  begin
    iPos2 := RPosEX(PathDelim, s, iPos - 1);
    Delete(s, iPos2, iPos - iPos2 + 3);
  end;
  Result := ExpandEnvironmentString(s);
  If Pos(':\\', Result) = 2 Then
    Delete(Result, 3, 1); // C:\\ << eines der beiden \ löschen
end;

procedure CreateDirsUTF8(s: string);
var
  sl: TStringListUTF8;
  sDir: string;
begin
  if s = '' then
    Exit;
  if s[1] = '.' then
    Exit; // Nur absolute Pfadangaben
  if DirectoryExistsUTF8Mm(s) then
    Exit;
  sl := TStringListUTF8.Create;
  sl.Delimiter := PathDelim;
  sl.StrictDelimiter := True; // Leerzeichen nicht als Trenner nehmen!
  sl.DelimitedText := s;
  sDir := sl[0];
  if (Length(sl[0]) = 2) and (Pos(':', sl[0]) = 2) then
    sl.Delete(0);
  if (sl[0] = '') and (sl[1] = '') then
  begin
    sl.Delete(0);
    sl.Delete(0);
    sDir := PathDelim + PathDelim;
  end;
  while sl.Count > 0 do
  begin
    sDir := sDir + PathDelim + sl[0];
    sl.Delete(0);
    if not DirectoryExistsUTF8Mm(sDir) then
      CreateDirUTF8Mm(sDir);
  end;
end;

function CopyFileUTF8(const SrcFilename, DestFilename: string;
  PreserveTime: boolean = False): boolean;
var
  SrcFS: TFileStreamUTF8;
  DestFS: TFileStreamUTF8;
begin
  try
    SrcFS := TFileStreamUTF8.Create(SrcFilename, fmOpenRead or fmShareDenyWrite);
    try
      DestFS := TFileStreamUTF8.Create(DestFilename, fmCreate);
      try
        DestFS.CopyFrom(SrcFS, SrcFS.Size);
      finally
        DestFS.Free;
      end;
      if PreserveTime then
        FileSetDateUTF8(DestFilename, FileGetDate(SrcFS.Handle));
    finally
      SrcFS.Free;
    end;
    Result := True;
  except
    Result := False;
  end;
end;


end.

