unit fMain;

{$mode objfpc}{$H+}

interface

uses
  LCLIntf, LCLType, SysUtils, Classes, Graphics, Controls, Forms, MmUtilk,
  StdCtrls, EditBtn, FileUtil, Spin, ComCtrls, ExtCtrls, Buttons, Menus,
  Registry, fABout, Windows, Clipbrd, uscaledpi;

const CountThreads = 7; // 1= 2 Threads 7= 8 Threads

type
  TIsReadyEvent = procedure(iInstance, iExitCode: Integer) of object;
  TErrorEvent = procedure(iInstance: Integer; sError: String) of object;

  TCompressThread = Class(TThread)
  Private
	  FFile : String;
	  FCompress: Integer;
	  FReduce: Boolean;
	  FInstance: Integer;
    FOnIsReady: TIsReadyEvent;
    FError: TErrorEvent;
    sError: String;
  Protected
	  Procedure Execute; Override;
    Procedure IsReady();
    Procedure Error();
  Public
	  Constructor Create(iInstance: Integer);
	  Destructor Destroy(); Override;
	  Procedure DoCompress(sFile: String; iCompress: Integer; bReduce: Boolean);
    property OnIsReady: TIsReadyEvent read FOnIsReady write FOnIsReady;
    property OnError: TErrorEvent read FError write FError;
  End;

  { TfrmMain }

  TfrmMain = class(TForm)
    btnEnde: TButton;
    btnStart: TButton;
    chkBMP: TCheckBox;
    chkSmall: TCheckBox;
    chkWithSubDir: TCheckBox;
    edCompress: TSpinEdit;
    edDir: TDirectoryEdit;
    Label1: TLabel;
    Label2: TLabel;
    lbMsg: TLabel;
    lstLog: TListBox;
    mnCypAll: TMenuItem;
    mnCpy1: TMenuItem;
    panMain: TPanel;
    pgBar: TProgressBar;
    popCpy: TPopupMenu;
    spView: TSpeedButton;
	  procedure btnEndeClick(Sender: TObject);
	  procedure btnStartClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
	  procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mnCpy1Click(Sender: TObject);
    procedure mnCypAllClick(Sender: TObject);
    procedure popCpyPopup(Sender: TObject);
    procedure spViewClick(Sender: TObject);
  private
	  { Private-Deklarationen }
    bWantClose: Boolean;
    slFileList: TStringList;
    sStartPath: String;
	  procedure WMSysCommand(VAR Message: TWMSysCommand); message WM_SYSCOMMAND;
	  Procedure ReadDir(sStartDir: String);
	  Procedure ReadFiles(sStartDir, sExtension: String);
    Procedure ComressFiles();
  public
	  { Public-Deklarationen }
    Procedure ThreadIsReady(iInstance, iExitCode: Integer);
    Procedure ThreadError(iInstance: Integer; sError: String);
  end;

var
  frmMain: TfrmMain;
  thArr: Array[0..CountThreads] Of TCompressThread;
  thStat: Array[0..CountThreads] Of Integer;  // Debug-Ausgabe um zu sehen warum welcher der Thread wo hängen bleibt
  thFile: Array[0..CountThreads] Of String;
  evArr: Array[0..CountThreads] Of Boolean;

implementation

uses BGRABitmap, BGRABitmapTypes;

{$R *.lfm}

procedure TfrmMain.WMSysCommand(var Message: TWMSysCommand);
begin
	Inherited;
	If Message.CmdType = $F200  Then // Info Copyright
	Begin
		If frmAbout = Nil Then
			frmAbout := TfrmAbout.Create(Self);
		frmAbout.ShowModal;
	End;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
Var Reg: TRegistry;
	 i: Integer;
begin
  ScaleDPI(Self); // 96 ist der DPI-Wert beim Entwurf des Formulars
	lbMsg.Caption := '  Copyright by MmVisual.   www.MmVisual.de';
  bWantClose := False;
  slFileList := TStringList.Create;
	Reg := TRegistry.Create;
	Try
		Reg.RootKey := HKEY_CURRENT_USER;
		If Reg.OpenKey('\Software\MmVisual\CompressJPG', False) Then
		Begin
			Try
				chkWithSubDir.Checked := Reg.ReadBool('WithSubDir');
				chkBMP.Checked := Reg.ReadBool('WithBMP');
				If Reg.ValueExists('MakeSmall') Then
					chkSmall.Checked := Reg.ReadBool('MakeSmall');
				edCompress.Value := Reg.ReadFloat('Compress');
				edDir.Text := Reg.ReadString('Dir');
			Finally
				Reg.CloseKey;
			End;
		End;
	Finally
		Reg.Free;
	End;
	Try
		If Not DirectoryExistsUTF8(edDir.Text) { *Converted from DirectoryExists* } Then
			edDir.Text := GetCurrentDirUTF8 { *Converted from GetCurrentDir* }Else If edDir.Text = '' Then
			edDir.Text := GetCurrentDirUTF8; { *Converted from GetCurrentDir* }
	Except
	End;
	AppendMenu(GetSystemMenu(Handle, False), MF_STRING, $F200, 'Copyright ...');
	For i := 0 To CountThreads Do
	Begin
		evArr[i] := False;
		thArr[i] := TCompressThread.Create(i);
    thArr[i].OnIsReady := @ThreadIsReady;
    thArr[i].OnError := @ThreadError;
	End;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
Var i: Integer;
begin
	For i := 0 To CountThreads Do
	Begin
    thArr[i].Terminate;
  End;
	For i := 0 To CountThreads Do
	Begin
    while (not thArr[i].Terminated) do
    Begin
      Application.ProcessMessages;
      Sleep(10);
    End;
  End;
  slFileList.Free;
end;

procedure TfrmMain.mnCpy1Click(Sender: TObject);
begin
  If lstLog.Count <= 0 Then Exit;
  Clipboard.AsText := lstLog.Items[lstLog.ItemIndex];
end;

procedure TfrmMain.mnCypAllClick(Sender: TObject);
begin
  Clipboard.AsText := lstLog.Items.Text;
end;

procedure TfrmMain.popCpyPopup(Sender: TObject);
begin
  mnCypAll.Enabled := lstLog.Count > 0;
  mnCpy1.Enabled := mnCypAll.Enabled And (lstLog.ItemIndex >= 0);
end;

procedure TfrmMain.spViewClick(Sender: TObject);
begin
  If spView.Down Then
  Begin
    Width := panMain.Width + 300;
  end Else Begin
    Width := panMain.Width;
  end;
end;

procedure TfrmMain.btnEndeClick(Sender: TObject);
begin
  bWantClose := True;
  if btnStart.Enabled or (btnEnde.Tag = 1) Then // Bei 2x Drücken EXIT der EXE
  	Close
  Else Begin
    lstLog.Items.Add('Abbruch');
    lstLog.ItemIndex := lstLog.Count - 1;
  End;
  btnEnde.Tag := 1;
end;

procedure TfrmMain.btnStartClick(Sender: TObject);
begin
  btnStart.Enabled := False;
  btnEnde.Tag := 0;
  lstLog.Clear;
	lbMsg.Caption := '';
  bWantClose := False;
	If Not DirectoryExistsUTF8(edDir.Text) { *Converted from DirectoryExists* } Then
	Begin
		lbMsg.Caption := 'Das Eingegeben Verzeichnis existiert nicht!';
		Exit;
	end;
	Screen.Cursor := crHourGlass;
  slFileList.Clear;
  pgBar.Position := 0;
  pgBar.Max := 100;
	Try
    sStartPath := IncludeTrailingPathDelimiter(edDir.Text);
		ReadDir(sStartPath);
    pgBar.Max := slFileList.Count;
    ComressFiles();
	Finally
		Screen.Cursor := crDefault;
    lbMsg.Caption := 'Fertig.';
    btnStart.Enabled := True;
	End;
end;

procedure TfrmMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
Var Reg: TRegistry;
begin
  CloseAction := caFree;
  bWantClose := True;
	Reg := TRegistry.Create;
	Try
		Reg.RootKey := HKEY_CURRENT_USER;
		If Reg.OpenKey('\Software\MmVisual\CompressJPG', True) Then
		Begin
			Reg.WriteBool('WithSubDir', chkWithSubDir.Checked);
			Reg.WriteBool('WithBMP', chkBMP.Checked);
			Reg.WriteBool('MakeSmall', chkSmall.Checked);
			Reg.WriteFloat('Compress', edCompress.Value);
			Reg.WriteString('Dir', edDir.Text);
			Reg.CloseKey;
		End;
	Finally
		Reg.Free;
	End;
end;

Procedure TfrmMain.ReadFiles(sStartDir, sExtension: String);
Var sr: TSearchRec;
Begin
	If FindFirstUTF8(sStartDir + '*.' + sExtension, faAnyFile - faDirectory, sr) = 0 Then
	Begin
	  Repeat
      If SameText(ExtractFileExt(sr.Name), '.' + sExtension) Then
        slFileList.Add(sStartDir + sr.Name);
    Until(FindNextUTF8(sr) <> 0); // Alle Bilder in Liste ausfnehmen
	  FindCloseUTF8(sr);
  End;
End;

Procedure TfrmMain.ComressFiles();
Var i, k: Integer;
	bDone: Boolean;
  sFileMsg: String;
Begin
  k := 0;
  While slFileList.Count > 0 Do
	Begin
    sFileMsg := '.' + Copy(slFileList[0], Length(IncludeTrailingPathDelimiter(sStartPath)), Length(slFileList[0]));
		lbMsg.Caption := 'Komprimiere Bild: ' + sFileMsg;
		Refresh;
		Application.ProcessMessages;
		bDone := False;
    Inc(k);
		While Not bDone and not bWantClose Do
    Begin
			For i := 0 To CountThreads Do // Der nächste freie Thread verwenden
				If Not evArr[i] Then
				Begin
          lstLog.Items.Add('F' + IntToStr(i) + '/' + IntToStr(k) + ': ' + sFileMsg);
          lstLog.ItemIndex := lstLog.Count - 1;
          thFile[i] := slFileList[0];
					thArr[i].DoCompress(slFileList[0], Round(edCompress.Value), chkSmall.Checked);
					bDone := True;
					Break;
				End;
      Application.ProcessMessages;
    end;
    slFileList.Delete(0);
    If bWantClose Then
      slFileList.Clear;
	End;
  bDone := False;
  lstLog.Items.Add('WaitThreads.');
  lstLog.ItemIndex := lstLog.Count - 1;
  While Not bDone and not bWantClose Do
  Begin // Warte bis alle bearbeitet wurden
    Application.ProcessMessages;
    k := -1;
    For i := 0 To CountThreads Do
      If (Not evArr[i]) Or (evArr[i] And (thStat[i] = 25)) Then
        Inc(k);
    If k = CountThreads Then
      bDone := True;
  end;
  lstLog.Items.Add('Finish.');
  lstLog.ItemIndex := lstLog.Count - 1;
End;

Procedure TfrmMain.ReadDir(sStartDir: String);
Var sr: TSearchRec;
Begin
	sStartDir := IncludeTrailingPathDelimiter(sStartDir);
	ReadFiles(sStartDir, 'jpg');
	If chkBMP.Checked Then
		ReadFiles(sStartDir, 'bmp');
	If chkWithSubDir.Checked Then
	Begin
		If FindFirstUTF8(sStartDir + '*',faDirectory, sr) = 0 Then
		Begin
			Repeat
				If (sr.Name <> '.') And (sr.Name <> '..')
						And ((sr.Attr And faDirectory) = faDirectory) Then
					ReadDir(sStartDir + sr.Name + PathDelim); // Unterverzeichnisse mit komprimieren
			Until(FindNextUTF8(sr) <> 0);
			FindCloseUTF8(sr);
		End;
	End;
	lbMsg.Caption := 'Fertig.';
End;

Procedure TfrmMain.ThreadIsReady(iInstance, iExitCode: Integer);
Begin
  evArr[iInstance] := False;
  pgBar.Position := pgBar.Position + 1;
  If iExitCode <> 25 Then
  Begin
    lstLog.Items.Add('E' + IntToStr(iInstance) + ' EC: ' + IntToStr(iExitCode));
    lstLog.ItemIndex := lstLog.Count - 1;
  end;
end;

Procedure TfrmMain.ThreadError(iInstance: Integer; sError: String);
Begin
  lbMsg.Caption := IntToStr(iInstance) + ': ' + sError;
  lstLog.Items.Add('E' + lbMsg.Caption);
  lstLog.ItemIndex := lstLog.Count - 1;
end;

Constructor TCompressThread.Create(iInstance: Integer);
Begin
	Inherited Create(False);
	Priority := tpLower;
	FreeOnTerminate := True;
	FInstance := iInstance;
  FOnIsReady := Nil;
End;

Destructor TCompressThread.Destroy();
Begin
	Inherited Destroy;
End;

Procedure TCompressThread.DoCompress(sFile: String; iCompress: Integer; bReduce: Boolean);
Begin
	FFile := sFile;
	FCompress := iCompress;
	FReduce := bReduce;
	evArr[FInstance] := True;
End;

Procedure TCompressThread.IsReady();
Begin
  if Assigned(FOnIsReady) then
    FOnIsReady(FInstance, thStat[FInstance]);
end;

Procedure TCompressThread.Error();
Begin
  if Assigned(FError) then
    FError(FInstance, sError);
end;

Procedure TCompressThread.Execute();
Var iNew: Int64;
	 ImgJ: TJPEGImage;
	 ImgB: Graphics.TBitmap;
	 s: TFileStreamUTF8;
	 hFil: HFILE;
	 sr: TSearchRec;
   sF: String;
   bWidth, bStart: Boolean;
   FBGRA: TBGRABitmap;
Begin
  bStart := False;
  Try
    while not Terminated Do
    Begin
      thStat[FInstance] := 0;
      if evArr[FInstance] Then
      Begin
        bStart := True;
        ImgJ := TJPEGImage.Create;
    	  ImgB := Graphics.TBitmap.Create;
        thStat[FInstance] := 1;
    	  Try
    		  If SameText(ExtractFileExt(FFile), '.jpg') Then
    		  Begin
    			  ImgJ.LoadFromFile(FFile); // JPG laden
    			  ImgB.Assign(ImgJ);
    		  End Else ImgB.LoadFromFile(FFile); // Bitmap laden
          thStat[FInstance] := 2;
          If (ImgB.Width = 0) or (ImgB.Height = 0) Then
          Begin
            bStart := False;
            Synchronize(@IsReady);
            thStat[FInstance] := 3;
            Continue;
          end;
          thStat[FInstance] := 4;
          Try
            If FileExistsUTF8(FFile + '_n') Then
              DeleteFileUTF8Mm(FFile + '_n');
            s := TFileStreamUTF8.Create(FFile + '_n', fmCreate);
            thStat[FInstance] := 5;
    		    Try
              bWidth := ImgB.Width > ImgB.Height;
              If ImgB.Width >= ImgB.Height Then
                bWidth := (ImgB.Width > 2520) // Querformat
              Else bWidth := (ImgB.Height > 2520); // Hochformat
              thStat[FInstance] := 6;
    			    If FReduce And bWidth Then
    			    Begin // Nur wenn größer als 2520 Pixel
                Try
                  thStat[FInstance] := 7;
                  FBGRA := TBGRABitmap.Create(ImgB);
                  thStat[FInstance] := 8;
                  Try
                    FBGRA.ResampleFilter := rfBestQuality;
                    BGRAReplace(FBGRA, FBGRA.Resample(ImgB.Width Div 2, ImgB.Height Div 2));
    				        ImgJ.Assign(FBGRA.Bitmap);
                    thStat[FInstance] := 9;
                  Finally
                    FBGRA.Free;
                  end;
                  thStat[FInstance] := 10;
                Except
                  On E:Exception Do
                  Begin
                    thStat[FInstance] := 11;
                    sError := E.Message;
                    Synchronize(@Error);
    				        ImgJ.Assign(ImgB);
                  End;
                end;
                thStat[FInstance] := 12;
                ImgJ.CompressionQuality := TJPEGQualityRange(100 - FCompress);
    				    ImgJ.SaveToStream(s);
                thStat[FInstance] := 13;
    			    End Else Begin
                thStat[FInstance] := 14;
    				    ImgJ.Assign(ImgB);
    				    ImgJ.CompressionQuality := TJPEGQualityRange(100 - FCompress);
    				    ImgJ.SaveToStream(s);
    			    End;
              thStat[FInstance] := 15;
    		    Finally
    			    s.Free;
    		    End;
    		    If FindFirstUTF8(FFile,faAnyFile,sr) { *Converted from FindFirst* } = 0 Then
    		    Begin
              thStat[FInstance] := 16;
              iNew := FileSize(FFile + '_n');
    			    If (iNew + 4095) < sr.Size Then
    			    Begin // Neueu Datei ist kleiner (+ ein Festplattencluster)
                thStat[FInstance] := 17;
    				    DeleteFileUTF8(FFile); { *Converted from DeleteFile* } // Originaldatei löschen
                sF := ChangeFileExt(FFile, '.jpg');
                thStat[FInstance] := 18;
    				    RenameFileUTF8Mm(FFile + '_n',sF); { *Converted from RenameFile* }
                hFil := FileOpenUTF8(sF, fmOpenReadWrite Or fmShareDenyNone);
                thStat[FInstance] := 19;
    				    Try // Datum gleich der Originaldatei + 2 Sekunden setzen, damit sie minimal neuer ist
    					    FileSetDateUTF8Mm(sF,DateTimeToFileDate(FileDateToDateTime(sr.Time) + (2.0 / 86400.0))); { *Converted from FileSetDate* }
                  thStat[FInstance] := 20;
    				    Finally
    					    FileClose(hFil); { *Converted from CloseHandle* }
    				    End;
                thStat[FInstance] := 21;
              End Else Begin
                thStat[FInstance] := 22;
    				    DeleteFileUTF8Mm(FFile + '_n'); { *Converted from DeleteFile* } // Neue Datei ist grösser >> löschen
    			    End;
              thStat[FInstance] := 23;
    			    FindCloseUTF8(sr); { *Converted from FindClose* }
    		    End;
          Except
            On E:Exception Do
            Begin
              thStat[FInstance] := 24;
              sError := E.Message;
              Synchronize(@Error);
              Synchronize(@IsReady);
              bStart := False;
            End;
          end;
        Finally
    		  ImgJ.Free;
    		  ImgB.Free;
          thStat[FInstance] := 25;
          Synchronize(@IsReady);
          bStart := False;
    	  End;
      End else Sleep(20);
      thStat[FInstance] := 100;
    end;
  Finally
    If bStart Then
      Synchronize(@IsReady);
  End;
  thStat[FInstance] := 101;
End;

end.
